! WHIZARD 1.13 Aug 06 2001
! 
! (C) 1999-2001 by Wolfgang Kilian <kilian@particle.uni-karlsruhe.de>
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module vamp_interface

  use kinds, only: default !NODEP!

  use monte_carlo_states, only: monte_carlo_state, mcs, mc_process_index
  use monte_carlo_states, only: write
  use monte_carlo_states, only: current_event, rng
  use monte_carlo_states, only: monte_carlo_input
  use monte_carlo_states, only: checksum, checksum_error
  use monte_carlo_states, only: mc_get_accuracy, mc_get_efficiency
!  use monte_carlo_states, only: mc_get_efficiency_vector
  use monte_carlo_states, only: mc_write_result, mc_read_result
  use monte_carlo_states, only: mc_write_header, mc_write_hline
  use cuts, only: read, write, apply, mpi90_broadcast
  use beams, only: structure_function
  use phase_space, only: event_fill
  use events, only: event, set_in_state, set_out_state
  use particles, only: particle_momentum, set
  use lorentz, only: array
  use diagnostics, only: msg_buffer, msg_level
  use diagnostics, only: msg_message, msg_result, msg_warning, msg_error
  use limits, only: WHIZARD_ROOT, VERSION_STRLEN
  use limits, only: DEFAULT_FILENAME, GRID_EXTENSION_LENGTH

  use process_interface, only: process_scatter, process_scatter_colored !NODEP!
  use vamp, only: vamp_grids, vamp_create_grids !NODEP!
  use vamp, only: vamp_read_grids, vamp_write_grids !NODEP!
  use vamp, only: vamp_copy_grids, vamp_discard_integrals !NODEP!
  use vamp, only: vamp_sample_grids !NODEP!
  use vamp, only: vamp_refine_grids, vamp_update_weights !NODEP!
  use vamp, only: vamp_next_event, vamp_probability, vamp_get_variance !NODEP!
  use tao_random_numbers, only: tao_random_state, tao_random_number !NODEP!
  use mpi90, only: mpi90_rank, mpi90_broadcast, vamp_broadcast_grids !NODEP!
  use file_utils, only: free_unit !NODEP!

  implicit none
  private

  public :: mc_integrate
  public :: mc_enable_grid
  public :: mc_generate 

contains

  subroutine mc_integrate (mcs, mci, filename, time_per_event)

    type(monte_carlo_state), intent(inout) :: mcs
    type(monte_carlo_input), intent(in) :: mci
    character(len=*), intent(in) :: filename
    real(kind=default), intent(out) :: time_per_event
    character(len=len(filename)+len(mcs%process_id_current)+1) :: &
         & filename_current

    type result
       real(kind=default) :: integral, error, chi2, efficiency, accuracy
    end type result

    integer :: proc_id
    type(result) :: best
    integer :: it_best, it_file

    call mpi90_rank (proc_id)
    mcs%it = 0
    mcs%it_vamp = 0
    filename_current = trim(filename)//"."//trim(mcs%process_id_current)
    best = result (0._default, 0._default, 0._default, &
         &         0._default, huge(1._default)) 

    call integrate_create_grids (mcs%weights(1,:), mcs%calls(2,1))
    mcs%checksum = checksum (mcs, 1)
    call integrate_read_grids (trim(filename_current), it_best, it_file)
    call mc_write_header (6, mcs%process_id_current)
    call integrate_fixed_weights (1)
    call integrate_variable_weights (2)
    call integrate_fixed_weights (3, time_per_event)
    call integrate_warmup (4)

  contains

    subroutine integrate_fixed_weights (pass, time_per_event)
      integer, intent(in) :: pass
      real(kind=default), intent(out), optional :: time_per_event
      integer :: time0, time1, tick_rate
      integer :: it_initial, it_final
      logical :: improved_grid
      real(kind=default) :: tmp
      character(len=80) :: buffer
      it_initial = mcs%it
      it_final   = it_initial + 1
      call vamp_copy_grids (mcs%g, mcs%g_best)
      call integrate_read_cuts (filename, pass)
      if (it_file < it_final) then
         call vamp_discard_integrals &
              & (mcs%g, mcs%calls(2,pass), stratified=mcs%stratified)
         if (pass==1) then
            buffer = "Preparing (fixed weights):"
         else
            buffer = "Integrating (fixed wgts.):"
         end if
         if (mcs%calls(1,pass)==1) then
            write(msg_buffer, "(1x,A,1x,I3,1x,A,1x,I10,1x,A)") &
                 & trim(buffer), mcs%calls(1,pass), "sample of", &
                 & mcs%calls(2,pass), "calls ..."
         else
            write(msg_buffer, "(1x,A,1x,I3,1x,A,1x,I10,1x,A)") &
                 & trim(buffer), mcs%calls(1,pass), "samples of", &
                 & mcs%calls(2,pass), "calls ..."
         end if
         call msg_message
         call system_clock (time0, tick_rate)
         call vamp_sample_grids &
              & (rng, mcs%g, mc_sample, mcs%calls(1, pass), &
              &  integral=mcs%integral, std_dev=mcs%error, avg_chi2=mcs%chi2, &
              &  history=mcs%history(mcs%it_vamp+1:), &
              &  histories=mcs%histories(mcs%it_vamp+1:,:))
         call system_clock (time1)
         mcs%actual_integration_done = .true.
         mcs%efficiency = mc_get_efficiency (mcs%g)
         mcs%accuracy = mc_get_accuracy &
              & (mcs%integral, mcs%error, product(mcs%calls(:,pass)))
         call integrate_update_best_grid (it_final, improved_grid, it_best)
         call mc_write_result &
              & (mcs%logtable(it_final), &
              &  it_final, product(mcs%calls(:,pass)), &
              &  mcs%integral, mcs%error, mcs%efficiency, mcs%accuracy, &
              &  use_efficiency=mcs%use_efficiency, improved=improved_grid, &
              &  chi2=mcs%chi2, iterations=mcs%calls(1,pass))
         call integrate_write_grids (trim(filename_current), it_final, it_best)
      else
         call msg_message(" Using grids and results from file:")
         time0 = 0
         time1 = 0
         tick_rate = 0
         improved_grid = .false.
      end if
      if (present(time_per_event)) then
         tmp = tick_rate * product(mcs%calls(:,pass)) * best%efficiency
         if (tmp > 0) then
            time_per_event = (time1 - time0) / tmp
         else
            time_per_event = 0
         end if
      end if
      mcs%it = it_final
      mcs%it_vamp = mcs%it_vamp + 1
      mcs%weights (mcs%it,:) = mcs%g%weights
      if (mcs%write_logfile)  call write (filename, mcs, mci)
      call msg_result (mcs%logtable(it_final))
      call mc_write_hline (6)
    end subroutine integrate_fixed_weights

    subroutine integrate_variable_weights (pass)
      integer, intent(in) :: pass
      integer :: it_initial, it_final, it_new, it
      logical :: first_old, first_new, sufficient, improved_grid
      character (len=80) :: buffer
      it_initial = mcs%it
      it_final = it_initial + mcs%calls(1,pass)
      if (it_file <= 1) then
         it_new = it_final - 1
      else if (it_file <= it_final) then
         it_new = it_final - it_file
      else
         it_new = 0
      end if
      first_old = .true.
      first_new = .true.
      sufficient = .false.
      call vamp_copy_grids (mcs%g, mcs%g_best)
      call integrate_read_cuts (filename, pass)
      call integrate_check_accuracy (sufficient)
      do it = it_initial+1, it_final
         if (it_file < it) then
            if (.not.sufficient) then
               if (first_new) then
                  buffer = "Adapting (variable wgts.):"
                  if (it_new==1) then
                     write (msg_buffer, "(1x,A,1x,I3,1x,A,1x,I10,1x,A)") &
                          & trim(buffer), it_new, "sample of", &
                          & mcs%calls(2,pass), "calls ..."
                  else
                     write (msg_buffer, "(1x,A,1x,I3,1x,A,1x,I10,1x,A)") &
                          & trim(buffer), it_new, "samples of", &
                          & mcs%calls(2,pass), "calls ..."
                  end if
                  call msg_message
                  call vamp_discard_integrals &
                       & (mcs%g, mcs%calls(2,pass), stratified=mcs%stratified)
                  first_new = .false.
               end if
               call vamp_refine_grids (mcs%g)
               call integrate_refine_weights
               call vamp_sample_grids &
                    & (rng, mcs%g, mc_sample, 1, &
                    &  integral=mcs%integral, &
                    &  std_dev=mcs%error, avg_chi2=mcs%chi2, &
                    &  history=mcs%history(mcs%it_vamp+1:), &
                    &  histories=mcs%histories(mcs%it_vamp+1:,:))
               mcs%actual_integration_done = .true.
               mcs%efficiency = mc_get_efficiency (mcs%g)
               mcs%accuracy = mc_get_accuracy &
                    & (mcs%integral, mcs%error, mcs%calls(2,pass))
               call integrate_update_best_grid (it, improved_grid, it_best)
               call integrate_write_grids (trim(filename_current), it, it_best)
            else
               improved_grid = .false.
            end if
            call mc_write_result &
                 & (mcs%logtable(it), it, mcs%calls(2,pass), &
                 &  mcs%integral, mcs%error, mcs%efficiency, mcs%accuracy, &
                 &  use_efficiency=mcs%use_efficiency, &
                 &  improved=improved_grid)
         else
            if (first_old) then
               call msg_message (" Using grids and results from file:")
               first_old = .false.
            end if
            improved_grid = .false.
         end if
         mcs%it = it
         mcs%it_vamp = mcs%it_vamp + 1
         mcs%weights (mcs%it,:) = mcs%g%weights
         if (mcs%write_logfile)  call write (filename, mcs, mci)
         call msg_result (mcs%logtable(it))
         call integrate_check_accuracy (sufficient)
      end do
      if (mcs%screen_results .and. proc_id==WHIZARD_ROOT) then
         if ((mcs%accuracy_goal > 0 .or. mcs%efficiency_goal < 100) &
              & .and. .not.sufficient)  call msg_warning &
              & (" Accuracy and/or efficiency goal not reached.")
         call mc_write_hline (6)
      end if
    end subroutine integrate_variable_weights

    subroutine integrate_check_accuracy (sufficient)
      logical, intent(inout) :: sufficient
      if (sufficient)  return
      if (mcs%accuracy < mcs%accuracy_goal &
           & .and. 100*mcs%efficiency > mcs%efficiency_goal) then
         call msg_message (" Accuracy and efficiency goal have been reached.")
         sufficient = .true.
      else if (mcs%accuracy < mcs%accuracy_goal &
           & .and. mcs%efficiency_goal == 100) then
         call msg_message (" Accuracy goal has been reached.")
         sufficient = .true.
      else if (mcs%accuracy_goal == 0 &
           & .and. 100*mcs%efficiency > mcs%efficiency_goal) then
         call msg_message (" Efficiency goal has been reached.")
         sufficient = .true.
      else
         sufficient = .false.
      end if
    end subroutine integrate_check_accuracy

    subroutine integrate_warmup (pass)
      integer, intent(in) :: pass
      integer :: it_initial, it_final
      logical :: improved_grid
      if (mcs%calls(1,pass) == 0)  return
      it_initial = mcs%it
      it_final = it_initial + 1
      if (it_file < it_final) then
         call vamp_copy_grids (mcs%g, mcs%g_best)
         call vamp_discard_integrals &
              & (mcs%g, mcs%calls(2,pass), stratified=.false.)
         call integrate_read_cuts (filename, pass)
         write (msg_buffer, "(1x,A,1x,I3,1x,A,1x,I10,1x,A)") &
              & "Maximum weight search:     ", &
              & mcs%calls(1,pass), "sample of", &
              & mcs%calls(2,pass), "calls ..."
         call msg_message
         call vamp_sample_grids (rng, mcs%g, mc_sample, 1)
         mcs%actual_integration_done = .true.
         mcs%efficiency = mc_get_efficiency (mcs%g)
         call integrate_update_best_grid &
              & (it_final, improved_grid, it_best, force=.true.)
         call mc_write_result &
              & (mcs%logtable(it_final), &
              &  it_final, product(mcs%calls(:,pass)), &
              &  mcs%integral, mcs%error, mcs%efficiency, mcs%accuracy, &
              &  use_efficiency=mcs%use_efficiency, improved=improved_grid, &
              &  chi2=mcs%chi2, iterations=mcs%calls(1,pass))
         call integrate_write_grids (trim(filename_current), it_final, it_best)
      else
         call msg_message (" Using grids from file:")
         improved_grid = .false.
      end if
      mcs%it = it_final
      mcs%it_vamp = mcs%it_vamp + 1
      mcs%weights (mcs%it,:) = mcs%g%weights
      if (mcs%write_logfile)  call write (filename, mcs, mci)
      call mc_write_result (6, &
           &  it_final, product(mcs%calls(:,pass)), &
           &  mcs%integral, mcs%error, mcs%efficiency, mcs%accuracy, &
           &  use_efficiency=mcs%use_efficiency, improved=improved_grid, &
           &  chi2=mcs%chi2, iterations=mcs%calls(1,pass))
      call mc_write_hline (6)
    end subroutine integrate_warmup

    subroutine integrate_create_grids (weights_in, calls)
      real(kind=default), dimension(:), intent(in) :: weights_in
      integer, intent(in) :: calls
      real(kind=default), dimension(size(weights_in)) :: weights
      real(kind=default), dimension(2,mcs%f%n_dim) :: region
      region(1,:) = 0;  region(2,:) = 1
      weights = weights_in
      if (any(weights < 0)) then
         call msg_error ("Negative initial channel weight encountered.  I'll set it to zero.")
         where (weights < 0)
            weights = 0
         end where
      end if
      if (all(weights == 0)) then
         call msg_error ("At least one initial weight must be positive.  I'll use equal weights.")
         weights = 1
      end if
      call vamp_create_grids &
           & (mcs%g, region, calls, weights, stratified=mcs%stratified)
      call vamp_create_grids &
           & (mcs%g_best, region, calls, weights, stratified=mcs%stratified)
    end subroutine integrate_create_grids

    subroutine integrate_read_grids (filename, it_best, it_file)
      character(len=*), intent(in) :: filename
      integer, intent(out) :: it_best, it_file
      integer :: it_dummy
      character(len=len(filename)+4) :: file_grc, file_grb
      logical :: grc_exist, grb_exist
      if (proc_id == WHIZARD_ROOT) then
         file_grc = filename // ".grc"
         file_grb = filename // ".grb"
         inquire (file=file_grc, exist=grc_exist)
         inquire (file=file_grb, exist=grb_exist)
         if (mcs%read_grids .and. (grb_exist .or. grc_exist)) then
            if (.not.grb_exist)  file_grb = file_grc
            if (.not.grc_exist)  file_grc = file_grb
            if (mcs%read_grids_force) then
               call mc_read_grids &
                    & (mcs%g_best, file_grb, it_dummy, it_best, &
                    &  best%integral, best%error, &
                    &  best%chi2, best%efficiency, best%accuracy, mcs%logtable)
               call mc_read_grids &
                    & (mcs%g, file_grc, it_dummy, it_file, &
                    &  mcs%integral, mcs%error, &
                    &  mcs%chi2, mcs%efficiency, mcs%accuracy, mcs%logtable)
            else
               call mc_read_grids &
                    & (mcs%g_best, file_grb, it_dummy, it_best, &
                    &  best%integral, best%error, &
                    &  best%chi2, best%efficiency, best%accuracy, &
                    &  mcs%logtable, mcs%checksum)
               call mc_read_grids &
                    & (mcs%g, file_grc, it_dummy, it_file, &
                    &  mcs%integral, mcs%error, &
                    &  mcs%chi2, mcs%efficiency, mcs%accuracy, &
                    &  mcs%logtable, mcs%checksum)
            end if
            call vamp_broadcast_grids (mcs%g, WHIZARD_ROOT)
            call vamp_broadcast_grids (mcs%g_best, WHIZARD_ROOT)
         else
            it_best = 0
            it_file = 0
         end if
      end if
      call mpi90_broadcast (best%integral, WHIZARD_ROOT)
      call mpi90_broadcast (best%error, WHIZARD_ROOT)
      call mpi90_broadcast (best%chi2, WHIZARD_ROOT)
      call mpi90_broadcast (best%efficiency, WHIZARD_ROOT)
      call mpi90_broadcast (best%accuracy, WHIZARD_ROOT)
      call mpi90_broadcast (it_best, WHIZARD_ROOT)
      call mpi90_broadcast (it_file, WHIZARD_ROOT)
    end subroutine integrate_read_grids

    subroutine integrate_write_grids (filename, it, it_best)
      character(len=*), intent(in) :: filename
      integer, intent(in) :: it, it_best
      integer, parameter :: itext_len = GRID_EXTENSION_LENGTH
      character(len=len(filename)+4) :: file_grc, file_grb
      character(len=len(filename)+3+itext_len) :: file_grx
      character(len=*), parameter :: &
           & itext_fmt = "(A,A,i" // char(ichar('0')+itext_len) // &
           &             "." // char(ichar('0')+itext_len) // ")"
      if (.not.mcs%write_grids) return
      if (proc_id == WHIZARD_ROOT) then
         file_grc = filename // ".grc"
         file_grb = filename // ".grb"
         call mc_write_grids (mcs%g_best, file_grb, it, it_best, &
              & best%integral, best%error, best%chi2, best%efficiency, &
              & best%accuracy, &
              & mcs%logtable(:it), trim(mcs%process_id_current), mcs%checksum)
         call mc_write_grids (mcs%g, file_grc, it, it, &
              & mcs%integral, mcs%error, mcs%chi2, mcs%efficiency, &
              & mcs%accuracy, &
              & mcs%logtable(:it), trim(mcs%process_id_current), mcs%checksum)
         if (mcs%write_all_grids) then
            write (file_grx,itext_fmt) filename, ".gr", it
            call mc_write_grids (mcs%g, file_grx, it, it, &
                 & mcs%integral, mcs%error, mcs%chi2, mcs%efficiency, &
                 & mcs%accuracy, &
                 & mcs%logtable(:it), trim(mcs%process_id_current), mcs%checksum)
         end if
      end if
    end subroutine integrate_write_grids

    subroutine integrate_update_best_grid (it, improved_grid, it_best, force)
      integer, intent(in) :: it
      logical, intent(out) :: improved_grid
      integer, intent(inout) :: it_best
      logical, intent(in), optional :: force
      logical :: check
      check = .true.
      if (present(force)) then
         if (force)  check = .false.
      end if
      if (check) then
         if (mcs%use_efficiency) then
            improved_grid = (mcs%efficiency > best%efficiency)
         else
            improved_grid = (mcs%accuracy < best%accuracy)
         end if
      else
         improved_grid = .true.
      end if
      if (improved_grid) then
         call vamp_copy_grids (mcs%g_best, mcs%g)
         best%integral = mcs%integral
         best%error = mcs%error
         best%chi2 = mcs%chi2
         best%efficiency = mcs%efficiency
         best%accuracy = mcs%accuracy
         it_best = it
      end if
    end subroutine integrate_update_best_grid

    subroutine integrate_refine_weights
      real(kind=default), dimension(size(mcs%g%weights)) :: weights
      real(kind=default) :: w
      integer :: i
      if (mcs%use_efficiency) then
         weights = mcs%g%weights &
              & * mc_get_efficiency_vector(mcs%g) ** mcs%weights_power
      else
         weights = mcs%g%weights &
              & * vamp_get_variance (mcs%g%grids) ** mcs%weights_power
      end if
      do i=1, mcs%f%n_groves
         w = sum(weights(mcs%f%grove(i)%first : mcs%f%grove(i)%last)) &
              &       / (mcs%f%grove(i)%last - mcs%f%grove(i)%first + 1)
         weights(mcs%f%grove(i)%first : mcs%f%grove(i)%last) = w
      end do
      call vamp_update_weights(mcs%g, weights)
    end subroutine integrate_refine_weights

    function mc_get_efficiency_vector (g) result(efficiency)
      type(vamp_grids), intent(in) :: g
      real(kind=default), dimension(size(g%grids)) :: efficiency
      where(g%num_calls >= 2 .and. g%grids%f_max>0)
         efficiency = g%grids%mu(1) / g%grids%f_max
      elsewhere
         efficiency = 0
      end where
    end function mc_get_efficiency_vector

    subroutine integrate_read_cuts (filename, pass)
      character(len=*), intent(in) :: filename
      integer, intent(in) :: pass
      character(len=max(len(filename), len(DEFAULT_FILENAME))+5) :: file
      character :: lext
      logical :: exist
      write(lext,'(i1)')  pass
      if (proc_id==WHIZARD_ROOT) then
         file = filename//".cut"//lext
         inquire (file=trim(file), exist=exist)
         if (.not.exist)  file = DEFAULT_FILENAME//".cut"//lext
         call read (trim(file), mcs%cc, trim(mcs%process_id_current), pass)
         if (mcs%screen_results) call write(6, mcs%cc, pass)
      end if
      call mpi90_broadcast(mcs%cc, WHIZARD_ROOT)
    end subroutine integrate_read_cuts


  end subroutine mc_integrate

  subroutine mc_write_grids &
       & (g, name, iterations, it, &
       &  integral, error, chi2, efficiency, accuracy, &
       &  logtable, process_id, checksum)
    type(vamp_grids), intent(in) :: g
    character(len=*), intent(in) :: name, process_id
    integer, intent(in) :: iterations, it
    real(kind=default), intent(in) :: integral, error, chi2
    real(kind=default), intent(in) :: efficiency, accuracy
    character(len=*), dimension(:), intent(in) :: logtable
    integer, intent(in) :: checksum
    integer :: u
    logical, parameter :: write_integrals = .true.
    u = free_unit()
    open(u, file=name, action="write", status="replace")
    write(u, "(1x,a)")   "WHIZARD version 1.13 (Aug 06 2001)"
    write(u, "(1x,a16,1x,a)")    "process_id     =", process_id
    write(u, "(1x,a16,1x,i15)")  "checksum       =", checksum
    write(u, "(1x,a16,1x,i4)")   "iterations     =", iterations
    write(u, "(1x,a16,1x,i4)")   "this_iteration =", it
    write(u, "(1x,a16,1x,e30.22)") "integral       =", integral
    write(u, "(1x,a16,1x,e30.22)") "error          =", error
    write(u, "(1x,a16,1x,e30.22)") "chi2           =", chi2
    write(u, "(1x,a16,1x,e30.22)") "efficiency     =", efficiency
    write(u, "(1x,a16,1x,e30.22)") "accuracy       =", accuracy
    write(u, "(1x,a)")   "begin logtable"
    call mc_write_result(u, iterations, logtable)
    write(u, "(1x,a)")   "end logtable"
    call vamp_write_grids(g, u, write_integrals)
    close(u)
  end subroutine mc_write_grids
  subroutine mc_read_grids &
       & (g, name, iterations, it, &
       &  integral, error, chi2, efficiency, accuracy, &
       &  logtable, checksum)
    type(vamp_grids), intent(inout) :: g
    character(len=*), intent(in) :: name
    integer, intent(inout) :: iterations, it
    real(kind=default), intent(inout) :: integral, error, chi2
    real(kind=default), intent(inout) :: efficiency, accuracy
    character(len=*), dimension(:), intent(inout) :: logtable
    integer, intent(in), optional :: checksum
    character(len=VERSION_STRLEN) :: version_string
    real :: version
    real, parameter :: current_version = 1.13
    character :: cdummy
    integer :: u, chsum
    logical :: ok, read_integrals
    read_integrals = .true.
    inquire(file=name, exist=ok)
    if (ok) then
       call msg_message (" Reading grid data from file " // name)
       u = free_unit()
       open(u, file=name, action="read", status="old")
       read(u, "(1x,a)") version_string
       if (version_string /= "WHIZARD version 1.13 (Aug 06 2001)") then
          call msg_warning (" Grid file has been generated by " // &
               &            trim(version_string))
          read(version_string, "(a15,1x,g5.2)") cdummy, version
          if (version < 1.12)  then
             call msg_warning (" Using compatibility hack for version < 1.12")
             read_integrals = .false.
          end if
       end if
       read(u, *) cdummy
       read(u, "(1x,a16,1x,i15)") cdummy, chsum
       read(u, "(1x,a16,1x,i4)")  cdummy, iterations
       read(u, "(1x,a16,1x,i4)")  cdummy, it
       read(u, "(1x,a16,1x,e30.22)") cdummy, integral
       read(u, "(1x,a16,1x,e30.22)") cdummy, error
       read(u, "(1x,a16,1x,e30.22)") cdummy, chi2
       read(u, "(1x,a16,1x,e30.22)") cdummy, efficiency
       read(u, "(1x,a16,1x,e30.22)") cdummy, accuracy
       if (present(checksum)) then
          if (chsum /= checksum) call checksum_error
       end if
       read(u, *)
       call mc_read_result(u, iterations, logtable)
       read(u, *)
       call vamp_read_grids(g, u, read_integrals)
       close(u)
    end if
  end subroutine mc_read_grids

  subroutine mc_enable_grid (mcs, g, safety_factor)
    type(monte_carlo_state), intent(inout) :: mcs
    type(vamp_grids), intent(inout) :: g
    real(kind=default), intent(in), optional :: safety_factor
    call vamp_copy_grids (mcs%g, g)
    if (present(safety_factor)) &
         & mcs%g%grids%f_max = mcs%g%grids%f_max * mcs%safety_factor
  end subroutine mc_enable_grid

  subroutine mc_generate(mcs, weight, probability, channel, excess)
    type(monte_carlo_state), intent(inout) :: mcs
    real(kind=default), intent(out), optional :: weight, probability
    integer, intent(in), optional :: channel
    real(kind=default), intent(out), optional :: excess
    real(kind=default), dimension(mcs%f%n_dim) :: x
    real(kind=default) :: wgt
    if (present(channel).and.present(probability)) then
       call vamp_next_event(x, rng, mcs%g%grids(channel), mc_sample, &
            &               wgt, channel, mcs%g%weights, mcs%g%grids)
       probability = wgt / mcs%g%grids(channel)%f_max
       if (present(weight)) weight = wgt
    else
       call vamp_next_event(x, rng, mcs%g, mc_sample, phi, excess=excess)
    end if
    call select_flavor_and_helicity &
         & (current_event, mcs%n_hel, mcs%n_col, &
         &  mcs%code_out, mcs%hel_state_out, mcs%col_flow, mcs%acl_flow, &
         &  rng)
    current_event%count = current_event%count + 1
  end subroutine mc_generate

  function phi (xi, channel_dummy) result (x)
    use kinds !NODEP!
    real(kind=default), dimension(:), intent(in) :: xi
    integer, intent(in) :: channel_dummy
    real(kind=default), dimension(size(xi)) :: x
    x = xi
  end function phi

  subroutine select_flavor_and_helicity &
       & (evt, n_hel, n_col, &
       &  flv_out, hel_out, col_flow, acl_flow, rng)
    type(event), intent(inout) :: evt
    integer, intent(in) :: n_hel, n_col
    integer, dimension(:,:), intent(in) :: flv_out, hel_out, col_flow, acl_flow
    type(tao_random_state), intent(inout) :: rng
    real(kind=default), dimension(evt%n_flv_out) :: rho_flv
    real(kind=default), dimension(evt%n_hel_out) :: rho_hel
    real(kind=default), dimension(evt%n_col_out) :: rho_col
    integer :: f, h, c, i
    real(kind=default) :: r
    rho_flv = sum(sum(evt%rho_out(:,:,:evt%n_flv_out), dim=2), dim=1)
    call tao_random_number (rng, r)
    call select (f, rho_flv, evt%n_flv_out, r)
    do i=1, evt%n_out
       call set (evt%prt(i), flv_out(i,f))
    end do
    if (n_hel > 0) then
       rho_hel = sum(evt%rho_out(:,:evt%n_hel_out,f), dim=1)
       rho_hel = rho_hel / sum(rho_hel)
       call tao_random_number (rng, r)
       call select (h, rho_hel, evt%n_hel_out, r)
       evt%hel_out = hel_out(:,h)
    else
       h = 1
       evt%hel_out = 0
    end if
    if (n_col > 0) then
       rho_col = evt%rho_out(:evt%n_col_out,h,f)
       rho_col = rho_col / sum(rho_col)
       call tao_random_number (rng, r)
       call select (c, rho_col, evt%n_col_out, r)
       evt%color_flow = col_flow(:,c)
       evt%anticolor_flow = acl_flow(:,c)
    else
       c = 1
       evt%color_flow = 0
       evt%anticolor_flow = 0
    end if
  contains
    subroutine select (k, rho, n, r)
      integer, intent(out) :: k
      real(kind=default), dimension(:), intent(in) :: rho
      integer, intent(in) :: n
      real(kind=default), intent(in) :: r
      integer :: i
      real(kind=default) :: rho_sum
      rho_sum = 0
      k = n
      LOOP: do i=1, n
         rho_sum = rho_sum + rho(i)
         if (r<rho_sum) then
            k = i
            exit LOOP
         end if
      end do LOOP
    end subroutine select
  end subroutine select_flavor_and_helicity

  function mc_sample (xi, weights, channel, grids) result (s)
    use vamp_grid_type !NODEP!
    real(kind=default), dimension(:), intent(in) :: xi
    real(kind=default), dimension(:), intent(in), optional :: weights
    integer, intent(in), optional :: channel
    type(vamp_grid), dimension(:), intent(in), optional :: grids
    real(kind=default) :: s

    call mc_sample_state (mcs(mc_process_index), s)

  contains

    subroutine mc_sample_state (mcs, s)
      type(monte_carlo_state), intent(inout) :: mcs
      real(kind=default), intent(out) :: s
      real(kind=default), dimension(mcs%f%n_dim, mcs%f%n_channels) :: rr
      real(kind=default), dimension(2) :: x_parton
      real(kind=default), dimension(mcs%f%n_channels) :: phase_space_factor
      real(kind=default) :: phase_space_volume, strfun_factor
      real(kind=default), dimension(mcs%f%n_channels) :: p_evt, g_xi
      logical, dimension(mcs%f%n_channels) :: active
      real(kind=default) :: dp
      integer :: ch
      logical :: ok
      current_event%process_id = mcs%process_id_current
      current_event%process_index = mc_process_index
      active = (mcs%g%num_calls >= 2)
      rr(:, channel) = xi
      s = 0
      if (all(mcs%beam%fixed_energy)) then
         strfun_factor = 1
         call set_in_state &
              & (current_event, mcs%code_in(:,1), mcs%par, mcs%sqrts)
         ok = .true.
      else
         do ch=1, mcs%n_channels
            rr(mcs%f%n_dim1+1:, ch) = rr(mcs%f%n_dim1+1:, channel)
         end do
         call structure_function &
              & (strfun_factor, x_parton, &
              &  rr(mcs%f%n_dim1+mcs%f%n_dim2+1:, channel), &
              &  mcs%sqrts, mcs%beam, mcs%code_in(:,1), ok)
         if (ok) call set_in_state&
              & (current_event, mcs%code_in(:,1), mcs%par, mcs%sqrts, &
              &  xi=x_parton, ok=ok)
      end if
      if (ok) then
         call set_out_state(current_event, mcs%code(mcs%n_in+1:,1), mcs%par)
         call event_fill(current_event, rr(:mcs%f%n_dim,:), mcs%f, &
              phase_space_factor, phase_space_volume, channel, active, ok)
         current_event%count_internal = current_event%count_internal + 1
      end if
      if (ok) call apply(mcs%cc, current_event, ok)
      if (ok) then
         where (.not.active)
            phase_space_factor = 0
         end where
         do ch=1, mcs%f%n_channels
            if (active(ch) .and. phase_space_factor(ch)>0) then
               p_evt(ch) = vamp_probability(grids(ch), rr(:,ch))
               g_xi(ch)  = p_evt(ch) / phase_space_factor(ch)
            else
               p_evt(ch) = 0
               g_xi(ch) = 0
            end if
         end do
         dp = dot_product(weights, g_xi)
         if (dp>0) then
            call scatter_set_probabilities &
                 & (mcs, current_event, trim(mcs%process_id_current), &
                 &  mcs%rho_in, mcs%rho_out, &
                 &  phase_space_volume * strfun_factor * p_evt(channel) / dp)
            call set_out_state (current_event, mcs%rho_out, &
                 &              mcs%n_col_out, mcs%n_hel_out, mcs%n_flv_out)
            s = current_event%mc_function_value
            if (any(current_event%rho_out<0)) then
               write (msg_buffer, "(1x,A,I10,1x,A)") &
                    & "Event #", current_event%count, "has negative entry in probability matrix."
               if (msg_level > 4) then
                  call msg_warning
                  print *, current_event%rho_out
               else
                  call msg_error
               end if
               current_event%rho_out = 0
            end if
         end if
      end if
      ! test !
      !    call write(6,current_event); stop
      ! test !
    end subroutine mc_sample_state

  end function mc_sample

  subroutine scatter_set_probabilities &
       & (mcs, evt, process_id, rho_in, rho_out, total_factor)
    type(monte_carlo_state), intent(in) :: mcs
    type(event), intent(inout) :: evt
    character(len=*), intent(in) :: process_id
    real(kind=default), dimension(:,:), intent(in) :: rho_in
    real(kind=default), dimension(:,:,:), intent(inout) :: rho_out
    real(kind=default), intent(in) :: total_factor
    real(kind=default), dimension(0:3, evt%n_tot) :: p
    real(kind=default) :: factor_tmp
    integer :: i
    if (evt%scattering) then
       do i=1,2
          p(:,i) = array(particle_momentum(evt%prt(-i)))
       end do
    else
       p(:,1) = array(particle_momentum(evt%prt(0)))
    end if
    do i=1, evt%n_out
       p(:,evt%n_in+i) = array(particle_momentum(evt%prt(i)))
    end do
    if (mcs%n_col > 0) then
       call process_scatter_colored (p, rho_in, rho_out, process_id)
    else
       call process_scatter (p, rho_in, rho_out(1,:,:), process_id)
    end if
    factor_tmp = sum(rho_out)
    if (factor_tmp /= 0) then
       rho_out = rho_out / factor_tmp
    end if
    evt%mc_function_value = factor_tmp * total_factor * evt%flux_factor
  end subroutine scatter_set_probabilities


end module vamp_interface
